from taggd.io.record import Record

class FASTARecord(Record):

    def __init__(self, object fqr):
        """Constructor. Represents a (annotation, seq) tuple as a Record."""
        Record.__init__(self)
        self.annotation = fqr[0]
        self.sequence = fqr[1]

    def add_tags(self, list added):
        #TODO this can be slow, use ''.{} format instead
        for k,v in added:
            self.annotation += " " + k + ":" + str(v)

    def unwrap(self):
        return (self.annotation, self.sequence)


    def __str__(self):
        fa_format = '>{header_comments}\n{sequence}\n'
        return fa_format.format(header_comments=self.annotation, sequence=self.sequence)





"""
Interface for writing/reading FASTAQ and SAM files
"""

import pysam as ps
import sys
import os
import taggd.io.fastq_utils as fu
from taggd.io.sam_record import *
from taggd.io.fasta_record import *
from taggd.io.fastq_record import *

# Global variables for match types.
cdef int FASTQ, FASTA, SAM, BAM

class ReadsReaderWriter():
    """
    Provides opening of a reads file, and writing to the same format to one or more
    specified files.
    """

    def __init__(self, str reads_infile_name):
        """Constructor"""

        self.file_type = -1
        self.infile_name = reads_infile_name
        self.infile = None
        self.infile_header = None

        # Supported file types
        global FASTQ
        global FASTA
        global SAM
        global BAM
        FASTQ, FASTA, SAM, BAM = range(4)

        cdef str suffix = self.infile_name.split(".")[-1].lower()
        if suffix == "fa" or suffix == "fasta":
            self.file_type = FASTA
        elif suffix == "fq" or suffix == "fastq":
            self.file_type = FASTQ
        elif suffix == "sam":
            self.file_type = SAM
        elif suffix == "bam":
            self.file_type = BAM
        else:
            raise ValueError("Unsupported reads file format!")

        # Read header.
        if self.file_type == SAM:
            self.infile = ps.AlignmentFile(self.infile_name, "r", check_header=True, check_sq=False)
            self.infile_header = self.infile.header
            self.infile.close()
        elif self.file_type == BAM:
            self.infile = ps.AlignmentFile(self.infile_name, "r", check_header=True, check_sq=False)
            self.infile_header = self.infile.header
            self.infile.close()

    def reader_open(self):
        """Opens the reads file using appropriate format."""

        # Open file.
        if self.file_type == FASTA or self.file_type == FASTQ:
            self.infile = fu.readfq(open(self.infile_name, "r"))
        elif self.file_type == SAM:
            self.infile = ps.AlignmentFile(self.infile_name, "r", check_header=True, check_sq=False)
        elif self.file_type == BAM:
            self.infile = ps.AlignmentFile(self.infile_name, "rb", check_header=True, check_sq=False)
        else:
            raise ValueError("Unsupported reads file format!")

        cdef object rec, last
        while True:
            last = None
            for orig in self.infile:
                if self.file_type == FASTA:
                    rec = FASTARecord(orig)
                    last = rec
                    yield rec
                    break
                elif self.file_type == FASTQ:
                    rec = FASTQRecord(orig)
                    last = rec
                    yield rec
                    break
                elif self.file_type == SAM or self.file_type == BAM:
                    rec = SAMRecord(orig)
                    last = rec
                    yield rec
                    break
                else:
                    raise ValueError("Unsupported reads file format!")
            if last == None:
                break

    def reader_close(self):
        """Closes the infile."""
        if self.infile != None:
            self.infile.close()
            self.infile = None

    def __exit__(self, type, value, tb):
        self.close_read()

    def get_format(self):
        """Returns the file format."""
        if self.file_type == FASTA:
            return "fa"
        if self.file_type == FASTQ:
            return "fq"
        if self.file_type == SAM:
            return "sam"
        if self.file_type == BAM:
            return "bam"
        return None

    def get_writer(self, str outfile_name):
        """
        Returns a writer.
        """
        if os.path.exists(outfile_name):
            os.remove(outfile_name)

        if self.file_type == FASTA or self.file_type == FASTQ:
            return open(outfile_name, "w")
        else:
            if self.infile_header == None:
                raise ValueError("Error: missing header in SAM/BAM file")
            if self.file_type == SAM:
                return ps.AlignmentFile(outfile_name, "wh", header=self.infile_header)
            if self.file_type == BAM:
                return ps.AlignmentFile(outfile_name, "wb", header=self.infile_header)
            else:
                return None

    def write_record(self, outfile, record):
        """Writes a record."""
        if self.file_type == FASTA:
            fu.writefa_record(outfile, record.unwrap())
        elif self.file_type == FASTQ:
            fu.writefq_record(outfile, record.unwrap())
        elif self.file_type == SAM:
            outfile.write(record.unwrap())
        elif self.file_type == BAM:
            outfile.write(record.unwrap())
        else:
            return


from multiprocessing import Process, Value, Lock

class Counter():
    def __init__(self, initval=0, typ='i'):
        self.val = Value(typ, initval)
        self.lock = Lock()

    def increment(self):
        with self.lock:
            self.val.value += 1

    def value(self):
        with self.lock:
            return self.val.value

    def add(self, val):
        with self.lock:
            self.val.value += val

    def __str__(self):
        return str(self.val.value)



""" 
Some functions to compute distance
between sequences
"""

cimport numpy as np
import numpy as np

cdef int hamming_distance(str seq1, str seq2, int limit=0):
    """
    Returns the Hamming distance between equal-length sequences.
    :param seq1: first sequence.
    :param seq2: second sequence.
    :param limit: max distance limit before aborting (returning limit + 1).
    :return: the edit distance.
    """
    cdef int i = 0
    cdef int sum = 0
    for i in xrange(len(seq1)):
        if seq1[i] != seq2[i]:
            sum += 1
        if limit > 0 and sum > limit:
            return limit + 1
    return sum


cdef int levenshtein_distance(str seq1, str seq2, int limit=0):
    """
    Returns the Levenshtein distance between two sequences. Lengths do no need to be equal, as indels are allowed.
    :param seq1: first sequence.
    :param seq2: second sequence.
    :param limit: max distance limit before aborting (returning limit + 1).
    :return: the edit distance.
    """
    cdef list one_ago = None
    cdef list this_row = range(1, len(seq2) + 1) + [0]
    cdef int x
    cdef int y
    cdef int del_cost
    cdef int add_cost
    cdef int sub_cost
    for x in xrange(len(seq1)):
        one_ago = this_row
        this_row = [0] * len(seq2) + [x + 1]
        for y in xrange(len(seq2)):
            del_cost = one_ago[y] + 1
            add_cost = this_row[y - 1] + 1
            sub_cost = one_ago[y - 1] + (seq1[x] != seq2[y])
            this_row[y] = min(del_cost, add_cost, sub_cost)
        if limit > 0 and x > limit and min(this_row) > limit:
            return limit + 1
    return this_row[len(seq2) - 1]


cdef list subglobal_distance(str s1, str s2):
    """
    Computes the edit distance for a sub-global (local?) alignment
    of a sequence s2 against a sequence s1.
    Mismatches and indels both score as 1. Overhanging parts of s1 do not count.
    :param s1: the longer (probe) sequence.
    :param s2: the shorter sought sequence.
    :return: [edit distance, last probe pos, s1 insertions, s2 insertions], where indels are estimates from a
    a minimal alignment, but indel counts are not necessarily minimal.
    """

    cdef int xLen = len(s1)
    cdef int yLen = len(s2)
    if xLen < yLen:
        raise ValueError("Sub-global edit distance is undefined for sequences where the probe is shorter than the aligned sequence.")

    cdef int x
    cdef int y

    cdef np.ndarray[np.uint32_t, ndim=2] d = np.empty([xLen+1, yLen+1], dtype=np.uint32)
    for x in xrange(0, xLen+1):
        d[x,0] = 0
    for y in xrange(1, yLen+1):
        d[0,y] = y # To ensure all of s2 is spanned.

    # Perform DP.
    for x in xrange(1, xLen+1):
        # Fill matrix.
        for y in xrange(1, yLen+1):
            d[x,y] = min( min(d[x-1,y]+1, d[x,y-1]+1), d[x-1,y-1] + int(s1[x-1] != s2[y-1]) )

    # Find min for sub-global alignment so that all of s2 is covered, but not necessarily all of s1 sequence.
    cdef int mini = 1000000
    cdef int iPos = 0
    cdef int i = xLen
    cdef int j
    while i > 0:
        if d[i,yLen] < mini:
            mini = d[i,yLen]
            iPos = i
        i -= 1

    # Compute backtracking for indels.
    cdef int s2ins = 0
    cdef int s1ins = 0

    # Last elements of alignment.
    i = iPos
    j = yLen

    # Backtrack.
    while (i > 0) and (j > 0):
        if d[i,j] == d[i-1,j-1] + int(s1[i-1] != s2[j-1]):
            # Subst.
            i -= 1
            j -= 1
        elif d[i,j] == d[i-1,j] + 1:
            # Add probe insertion. NOTE: Heavier weighting since this elif appears before below elif.
            s1ins += 1
            i -= 1
        else:
            # Add sought insertion.
            s2ins += 1
            j -= 1

    return [mini, iPos-1, s1ins, s2ins]   # Indexed from 1...



#!/usr/bin/env python
"""
Starts and runs a TagGD demultiplexing job. Type -h for help.
\n
See https://github.com/JoelSjostrand/taggd/wiki for manual and the most recent source code.
\n
See http://www.plosone.org/article/info%3Adoi%2F10.1371%2Fjournal.pone.0057521
for the peer-reviewed reference to the program.
"""


import taggd.core.demultiplex as deplex

if __name__ == "__main__":
    deplex.main()






class Record(object):
    """
    Shorthand for a record. To keep things simple, we use a single object
    regardless of file format, and let util classes deal with conversions.
    """

    def __init__(self):
        self.annotation = None
        self.sequence = None
        self.attributes = dict()

    def unwrap(self):
        """Unwraps an object to its proper original representation."""
        return None

    def add_tags(self, list added):
        """Adds tags provided as a list of (name, attr) pairs."""
        return

    def __str__(self):
        return str(self.annotation) + "\t" + str(self.sequence) + "\t" + str(self.attributes)


cdef class Barcode:
    cdef public str sequence
    cdef public list attributes

cpdef dict read_barcode_file(str infile_path)

cpdef int estimate_min_edit_distance(dict true_barcodes, int max_iters)


"""
Contains utilities for handling barcodes.
"""

from taggd.misc.distance_metrics cimport *
import random

cdef class Barcode:
    """
    Holds a Spatial Transcriptomics barcode sequence and feature coordinates.
    """

    def __cinit__(self, str seq_, list attributes_):
        self.sequence = seq_
        self.attributes = attributes_


cpdef dict read_barcode_file(str infile_path):
    """
    Reads a barcode ("chip") file and returns the barcodes.
    :param infile_path: chip file.
    :return: a dictionary with barcode sequences as keys and Barcode instances as values.
    """
    cdef dict res_dict = {}
    cdef str line
    cdef list tmp
    cdef str seq
    cdef int length = 0
    with open(infile_path) as infile:
        for line in infile:
            tmp = line.strip().split()
            seq = tmp[0]
            if length == 0:
                length = len(seq)
            if len(seq) != length:
                raise ValueError("Barcode file incorrect, varying lengths among barcodes")
            if seq in res_dict:
                raise ValueError("Barcode file incorrect, duplicate barcode: " + seq)
            res_dict[tmp[0]] = Barcode(tmp[0], tmp[1:])
    return res_dict

cpdef int estimate_min_edit_distance(dict true_barcodes, int max_iters):
    """
    Reads a barcodes dict and estimates the minimum edit distance
    by comparing a certain number of pairs.
    """

    cdef int length = 0
    cdef int min_dist = 1000000000
    cdef str ln
    cdef list tmp
    cdef str seq

    # Compute minimum edit distance.
    cdef list seqslist = true_barcodes.keys()
    random.shuffle(seqslist)
    cdef int i
    cdef int j
    cdef int dist
    cdef int iter = 0
    for i in xrange(len(seqslist)):
        for j in xrange(int(i+1), len(seqslist)):
            dist = hamming_distance(seqslist[i], seqslist[j], min_dist)
            if dist < min_dist:
                min_dist = dist
            iter += 1
            if iter >= max_iters:
                return min_dist
            #print str(i) + " vs " + str(j) + ": " + str(dist)
    return min_dist


"""
Contains utilities for working with k-mer chunks of the barcodes or other sequences.
"""
from cpython cimport bool

cdef list get_kmers_dicts(list seqs, int k, bool round_robin=False, int slider_increment=1):
    """
    Returns dictionaries for kmers of a list of sequences.
    The last kmer is always included, irrespective of slider increment.
    :param seqs: the sequences.
    :param k: the k-mer length.
    :param round_robin: if to treat the sequence as circular.
    :return: 1) a dictionary: seq -> set of kmer's of seq, and
             2) a dictionary: kmer -> dictionary of seqs holding kmer -> list of offsets of kmer in seq
    """
    cdef dict seq2kmer = {}
    cdef dict kmer2seq = {}
    cdef str seq
    cdef str seqq
    cdef set l
    cdef str kmer
    cdef int i
    for seq in seqs:
        l = set()
        seq2kmer[seq] = l
        seqq = seq
        if round_robin:
            seqq = seq + seq[0:(k-1)]
        for i in xrange(0, len(seqq)-k+1, slider_increment):
            kmer = seqq[i:i+k]
            l.add(kmer)
            if not kmer in kmer2seq:
                kmer2seq[kmer] = dict()
            if not seq in kmer2seq[kmer]:
                kmer2seq[kmer][seq] = list()
            kmer2seq[kmer][seq].append(i)
        # Special treatment of last in case we would skip it during incrementation
        if len(seqq) % slider_increment != 0:
            i = len(seqq)-k
            kmer = seqq[i:len(seqq)]
            l.add(kmer)
            if not kmer in kmer2seq:
                kmer2seq[kmer] = dict()
            if not seq in kmer2seq[kmer]:
                kmer2seq[kmer][seq] = list()
            kmer2seq[kmer][seq].add(i)
    return [seq2kmer, kmer2seq]


cdef list get_kmers(str seq, int k, bool round_robin=False, int slider_increment=1):
    """
    Returns the kmers of a sequence as a list of kmer-offset tuples. The last kmer will always be included
    irrespective of the slider increment.
    :param seq: sequence.
    :param k: kmer length.
    :param round_robin: if to treat the sequence as circular.
    :return: the kmers.
    """
    cdef list l = list()
    cdef str seqq = seq
    if round_robin:
        seqq = seq + seq[0:(k-1)]
    cdef str kmer
    cdef int i
    for i in xrange(0, len(seqq)-k+1, slider_increment):
        kmer = seqq[i:i+k]
        l.append((kmer, i))
    # Special treatment of last
    if len(seqq) % slider_increment != 0:
        i = len(seqq)-k
        kmer = seqq[i:len(seqq)]
        l.append((kmer,i))
    return l


from taggd.io.record import Record

class FASTQRecord(Record):

    def __init__(self, object fqr):
        """Constructor. Represents a (annotation, seq, qual) tuple as a Record."""
        Record.__init__(self)
        self.annotation = fqr[0]
        self.sequence = fqr[1]
        self.attributes["quality"] = fqr[2]

    def add_tags(self, list added):
        for k,v in added:
            self.annotation += " " + k + ":" + str(v)

    def unwrap(self):
        return (self.annotation, self.sequence, self.attributes["quality"])

    def __str__(self):
        fq_format = '@{header_comments}\n{sequence}\n+\n{quality}\n'
        return fq_format.format(header_comments=self.annotation, 
                                sequence=self.sequence, quality=self.attributes["quality"])


import pysam
from taggd.io.record import Record

class SAMRecord(Record):

    def __init__(self, object alseq):
        """Constructor. Represents a pysam.AlignedSequence as a Record."""
        Record.__init__(self)
        self.annotation = str(alseq.query_name)
        self.sequence = str(alseq.query_sequence)
        self.attributes["flag"] = alseq.flag
        self.attributes["reference_id"] = alseq.reference_id
        self.attributes["reference_start"] = alseq.reference_start
        self.attributes["mapping_quality"] = alseq.mapping_quality
        self.attributes["cigar"] = alseq.cigar
        self.attributes["next_reference_id"] = alseq.next_reference_id
        self.attributes["reference_start"] = alseq.reference_start
        self.attributes["template_length"] = alseq.template_length
        self.attributes["query_qualities"] = alseq.query_qualities
        self.attributes["tags"] = alseq.tags

    def add_tags(self, list added):
        self.attributes["tags"] += added

    def unwrap(self):
        cdef object a = pysam.AlignedSegment()
        a.query_name = self.annotation
        a.flag = self.attributes["flag"]
        a.reference_id = self.attributes["reference_id"]
        a.reference_start = self.attributes["reference_start"]
        a.mapping_quality = self.attributes["mapping_quality"]
        a.cigar = self.attributes["cigar"]
        a.next_reference_id = self.attributes["reference_id"]
        a.next_reference_start = self.attributes["reference_start"]
        a.template_length = self.attributes["template_length"]
        a.query_sequence = self.sequence
        a.query_qualities = self.attributes["query_qualities"]
        a.tags = self.attributes["tags"]
        return a


"""
Main functions for barcodes, like search and find matches
"""
cimport taggd.misc.kmer_utils as ku
cimport taggd.misc.distance_metrics as dm
from cpython cimport bool

# User options
cdef dict true_barcodes
cdef int k
cdef int max_edit_distance
cdef str metric
cdef int slider_increment
cdef int pre_overhang
cdef int post_overhang
cdef bool no_offset_speedup

# Dictionary
cdef dict kmer2seq = None

# Metrics
cdef int SUBGLOBAL
cdef int LEVENSHTEIN
cdef int HAMMING
cdef int metric_choice


def init(dict true_barcodes_,
        int k_,
        int max_edit_distance_,
        str metric_,
        int slider_increment_,
        int pre_overhang_,
        int post_overhang_,
        bool no_offset_speedup_):
    """
    Initializes settings (global variables).
    :param true_barcodes_: true barcodes dict to attributes.
    :param k_: the k-mer length used in the search algorithm.
    :param max_edit_distance_: max allowed distance.
    :param metric_: distance metric to use, Subglobal, Levenshtein or Hamming.
    :param slider_increment_: the number of increments for kmer search. The default, 0 sets it to k.
    :param pre_overhang: pre-read flanking bases.
    :param post_overhang: post-read flanking bases.
    """

    global true_barcodes
    true_barcodes = true_barcodes_
    global k
    k = k_
    global max_edit_distance
    max_edit_distance = max_edit_distance_
    global metric
    metric = metric_
    global slider_increment
    slider_increment = slider_increment_
    global pre_overhang
    pre_overhang = pre_overhang_
    global post_overhang
    post_overhang = post_overhang_
    global no_offset_speedup
    no_offset_speedup = no_offset_speedup_

    # Create k-mer mappings with ALL kmers
    global kmer2seq
    seq2kmer, kmer2seq = ku.get_kmers_dicts(true_barcodes.keys(), k, False, 1)

    # Metrics
    global SUBGLOBAL
    global LEVENSHTEIN
    global HAMMING
    SUBGLOBAL, LEVENSHTEIN, HAMMING = range(0, 3)
    global metric_choice
    if metric == "Subglobal":
        metric_choice = SUBGLOBAL
    elif metric == "Levenshtein":
        metric_choice = LEVENSHTEIN
    elif metric == "Hamming":
        metric_choice = HAMMING
    else:
        raise ValueError("Invalid distance metric specified")


cdef dict get_candidates(str read_barcode):
    """
    Returns candidate barcodes for a read barcode.
    :param read_barcode: read barcode.
    :return: the candidates as a dict with <barcode, read kmer hits>.
    """
    cdef dict candidates = dict()
    cdef list kmers_offsets = ku.get_kmers(read_barcode, k, False, slider_increment)
    cdef str kmer
    cdef int offset
    cdef dict hits
    cdef str hit
    cdef list hit_offsets
    cdef int hit_offset
    cdef int penalty = 0
    cdef int min_penalty = 0

    # NON-OPTIMIZED CASE
    # For each kmer in read (typically incremented by k positions at a time).
    if no_offset_speedup:
        for kmer, offset in kmers_offsets:
            try:
                hits = kmer2seq[kmer]
            except KeyError:
                continue
            # For each true barcode containing read's kmer.
            for hit, hit_offsets in hits.iteritems():
                candidates[hit] = 0
        return candidates

    # OPTIMIZED CASE
    # For each kmer in read (typically incremented by k positions at a time).
    for kmer, offset in kmers_offsets:
        # Obtain all the barcodes for a given kmer
        try:
            hits = kmer2seq[kmer]
        except KeyError:
            continue

        # For each true barcode containing read's kmer.
        for hit, hit_offsets in hits.iteritems():
            min_penalty = 100000000
            # For each position that kmer occurred in the true barcode.
            for hit_offset in hit_offsets:
                # Kmer may be shifted overhang positions without penalty, due to subglobal alignment.
                penalty = max(0, abs(offset - hit_offset) - pre_overhang - post_overhang)
                if penalty < min_penalty:
                    min_penalty = penalty
            #catching KeyError is faster than makin look-ups
            try:
                candidates[hit] = max(min_penalty, candidates[hit])
            except KeyError:
                candidates[hit] = min_penalty

    # Assure we remove the kmers
    del kmers_offsets
    
    # Clear out all candidates with a forced offset penalty greater than the max edit distance:
    for hit in candidates.keys():
        if candidates[hit] > max_edit_distance:
            del candidates[hit]
    return candidates


cdef list get_distances(str read_barcode, dict candidates):
    """
    Returns all qualified hits ordered by distance.
    :param read_barcode: the read barcode.
    :param candidates: the candidate dictionary.
    :return: a list of tuples, (barcode,distance,lastpos,ins,del).
    """
    cdef list qual_hits = []
    cdef str candidate = None
    cdef int penalty = 0
    cdef int dist = 0
    cdef int max_lim = 0
    cdef int read_last_pos = -1
    cdef int a = -1
    cdef int b = -1

    append = qual_hits.append
    for candidate, penalty in candidates.iteritems():
        if metric_choice == SUBGLOBAL:
            dist, read_last_pos, a, b = dm.subglobal_distance(read_barcode, candidate)
        elif metric_choice == LEVENSHTEIN:
            # Account for the added overhang!
            # Note: This does NOT equate subglobal and may miss cases subglobal would catch!
            max_lim = max_edit_distance + max(0, len(read_barcode) - len(candidate))
            dist = dm.levenshtein_distance(read_barcode, candidate, max_lim)
            read_last_pos = -1
            a = -1
            b = -1
        elif metric_choice == HAMMING:
            dist = dm.hamming_distance(read_barcode, candidate, max_edit_distance)
            read_last_pos = -1
            a = -1
            b = -1
        else:
            raise ValueError("Invalid distance metric specified")
        
        if dist <= max_edit_distance:
            append((candidate, dist, read_last_pos, a, b))
                
    return qual_hits


cdef list get_top_hits(list qual_hits):
    """
    Returns the top hits.
    :param qual_hits: the list of qualified hits.
    :return: the top hits.
    """
    if len(qual_hits) == 0:
        return None
    
    #TODO this can be optimized
    
    # Find smallest.
    cdef int mini = 10000000
    cdef str cand = None
    cdef int dist = 0
    cdef int last_pos = -1
    cdef int a = -1
    cdef int b = -1
    for cand, dist, last_pos, a, b in qual_hits:
        if dist < mini:
            mini = dist
    
    # Filter out the smallest
    cdef list top_hits = []
    append = top_hits.append
    for cand, dist, last_pos, a, b in qual_hits:
        if dist == mini:
            append((cand, dist, last_pos, a, b))
    return top_hits



"""
Some utilities to write/read
fasta/q files
"""

def coroutine(func):
    """
    Coroutine decorator, starts coroutines upon initialization.
    """
    def start(*args, **kwargs):
        cr = func(*args, **kwargs)
        cr.next()
        return cr
    return start

def readfq(fp): # this is a generator function
    """
    Heng Li's fasta/fastq reader function.
    """
    cdef str last = None # this is a buffer keeping the last unprocessed line
    cdef str l, name, seq
    cdef list seqs
    cdef int leng
    while True: # mimic closure; is it a bad idea?
        if not last: # the first record or a record following a fastq
            for l in fp: # search for the start of the next record
                if l[0] in '>@': # fasta/q header line
                    last = l[:-1] # save this line
                    break
        if not last: break
        #name, seqs, last = last[1:].partition(" ")[0], [], None
        name, seqs, last = last[1:], [], None
        for l in fp: # read the sequence
            if l[0] in '@+>':
                last = l[:-1]
                break
            seqs.append(l[:-1])
        if not last or last[0] != '+': # this is a fasta record
            yield name, ''.join(seqs), None # yield a fasta record
            if not last: break
        else: # this is a fastq record
            seq, leng, seqs = ''.join(seqs), 0, []
            for l in fp: # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(seq):  # have read enough quality
                    last = None
                    yield name, seq, ''.join(seqs)  # yield a fastq record
                    break
            if last:  # reach EOF before reading enough quality
                yield name, seq, None  # yield a fasta record instead
                break


@coroutine
def writefq(fp):  # This is a coroutine
    """
    Fastq writing generator sink.
    Send a (header_comments, sequence, quality) triple to the instance to write it to
    the specified file pointer.
    """
    cdef str fq_format = '@{header_comments}\n{sequence}\n+\n{quality}\n'
    cdef object record
    cdef str read
    try:
        while True:
            record = yield
            read = fq_format.format(header_comments=record[0], sequence=record[1], quality=record[2])
            fp.write(read)

    except GeneratorExit:
        return


@coroutine
def writefa(fp):  # This is a coroutine
    """
    Fasta writing generator sink.
    Send a (header_comments, sequence) double to the instance to write it to
    the specified file pointer.
    """
    cdef str fa_format = '>{header_comments}\n{sequence}\n'
    cdef object record
    cdef str read
    try:
        while True:
            record = yield
            read = fa_format.format(header_comments=record[0], sequence=record[1])
            fp.write(read)

    except GeneratorExit:
        return

def writefq_record(fp, record):
    """
    Send a (header_comments, sequence, quality) triple to the instance to write it to
    the specified file pointer.
    """
    cdef str fq_format = '@{header_comments}\n{sequence}\n+\n{quality}\n'
    cdef str read = fq_format.format(header_comments=record[0], sequence=record[1], quality=record[2])
    fp.write(read)


def writefa_record(fp, record):
    """
    Send a (header_comments, sequence) double to the instance to write it to
    the specified file pointer.
    """
    cdef str fa_format = '>{header_comments}\n{sequence}\n'
    cdef str read = fa_format.format(header_comments=record[0], sequence=record[1])
    fp.write(read)


cpdef double probability_to_phred(double p)

cpdef double phred_to_probability(double q)

cpdef str sanger_phred_to_ascii(list phred)

cpdef list sanger_ascii_to_phred(str ascii_seq)

cpdef list sanger_ascii_to_probability(str ascii_seq)

cpdef list illumina_1_3_ascii_to_phred(str ascii_seq)

cpdef list illumina_1_3_ascii_to_probability(str ascii_seq)

cpdef list illumina_1_8_casava_ascii_to_phred(str ascii_seq)

cpdef list illumina_1_8_casava_ascii_to_probability(str ascii_seq)

cpdef list solid_ascii_to_phred(str ascii_seq)

cpdef list solid_ascii_to_probability(str ascii_seq)


"""
Main functions for carrying out the demultiplexing with
multithreading.
"""
import os
import time
import multiprocessing as mp
import taggd.core.match as match
cimport taggd.core.match as match
import taggd.core.match_type as match_type
cimport taggd.core.match_type as match_type
import taggd.core.demultiplex_record_functions as demulti
import taggd.io.reads_reader_writer as rw
from taggd.misc.counter import *
from cpython cimport bool

# Global variables for user options
cdef dict true_barcodes
cdef str reads_infile
cdef str outfile_prefix
cdef int max_edit_distance
cdef bool no_multiprocessing
cdef bool only_output_matched
cdef int chunk_size
cdef int start_position
cdef int barcode_length

# Reader-writer
cdef object re_wr
cdef object f_match = None
cdef object f_res = None
cdef object f_ambig = None
cdef object f_unmatch = None


# Global variables for stats.
cdef object stats_total_reads
cdef object stats_total_reads_wr
cdef object stats_unmatched
cdef object stats_perfect_matches
cdef object stats_imperfect_unambiguous_matches
cdef object stats_imperfect_ambiguous_matches  # Non-unique
cdef list stats_edit_distance_counts
cdef int start_time

# Threading objects
cdef object manager


def init(dict true_barcodes_,
         str reads_infile_,
         str outfile_prefix_,
         int start_position_,
         int max_edit_distance_,
         bool no_multiprocessing_,
         bool only_output_matched_,
         int chunk_size_,
         int mp_chunk_size_):
    """
    Initializes settings (global variables).
    :param true_barcodes_: true barcodes.
    :param reads_infile_: input file with read sequences (containing read barcodes).
    :param outfile_prefix_: output file prefix.
    :param max_edit_distance_: Max edit distance.
    :param no_multiprocessing_: No multiprocessing.
    :param only_output_matched_: Only matched file out.
    """

    global true_barcodes
    true_barcodes = true_barcodes_
    global reads_infile
    reads_infile = reads_infile_
    global outfile_prefix
    outfile_prefix = outfile_prefix_
    global start_position
    start_position = start_position_
    global barcode_length
    barcode_length = len(true_barcodes.keys()[0])
    global max_edit_distance
    max_edit_distance = max_edit_distance_
    global no_multiprocessing
    no_multiprocessing = no_multiprocessing_
    global only_output_matched
    only_output_matched = only_output_matched_
    global chunk_size
    chunk_size = chunk_size_
    global mp_chunk_size
    mp_chunk_size = mp_chunk_size_

    # Reader writer
    global re_wr
    re_wr = rw.ReadsReaderWriter(reads_infile)

    # Stats
    global stats_total_reads
    stats_total_reads = Counter(0)
    global stats_total_reads_wr
    stats_total_reads_wr = Counter(0)
    global stats_unmatched
    stats_unmatched = Counter(0)
    global stats_perfect_matches
    stats_perfect_matches = Counter(0)
    global stats_imperfect_unambiguous_matches
    stats_imperfect_unambiguous_matches = Counter(0)
    global stats_imperfect_ambiguous_matches
    stats_imperfect_ambiguous_matches = Counter(0)
    global stats_edit_distance_counts
    stats_edit_distance_counts = []
    cdef int i
    for i in xrange(max_edit_distance+1):
        stats_edit_distance_counts.append(Counter(0))
    global start_time
    start_time = time.time()

    # Threading.
    global manager
    manager = mp.Manager()


def demultiplex():
    """
    Demultiplexes the contents of a reads file.
    It tries to find all the matches using the reads and barcodes
    present in global variables and it will write the results
    to different files.
    NOTE that this functions expects init() to have been called before.
    """

    # Open files.
    __open_files()

    # Demultiplex.
    cdef list chunk = list()
    cdef object rec

    # Process the reads in chunks either parallel or in single thread mode
    # Reference to function append to avoid overhead.
    append = chunk.append
    for rec in re_wr.reader_open():
        append(rec)
        stats_total_reads.increment()
        if stats_total_reads.value() % chunk_size == 0:
            if no_multiprocessing:
                __demultiplex_linearly_chunk(chunk)
            else:
                __demultiplex_mp_chunk(chunk)
            del chunk[:]
    # Process the rest if applies
    if no_multiprocessing:
        __demultiplex_linearly_chunk(chunk)
    else:
        __demultiplex_mp_chunk(chunk)
    del chunk[:]

    # Close all files.
    __close_files()


def __open_files():
    """
    Opens files for writing.
    """
    global f_match
    f_match = re_wr.get_writer(outfile_prefix + "_matched." + re_wr.get_format())
    if not only_output_matched:
        f_res_name = outfile_prefix + "_results.tsv"
        if os.path.exists(f_res_name):
            os.remove(f_res_name)
        global f_res
        f_res = open(f_res_name, 'w')
        f_res.write(match.get_match_header() + "\n")
        f_res.flush()
        global f_ambig
        f_ambig = re_wr.get_writer(outfile_prefix + "_ambiguous." + re_wr.get_format())
        global f_unmatch
        f_unmatch = re_wr.get_writer(outfile_prefix + "_unmatched." + re_wr.get_format())


def __close_files():
    """
    Closes output files.
    """
    re_wr.reader_close()
    f_match.close()
    if not only_output_matched:
        f_res.close()
        f_ambig.close()
        f_unmatch.close()


def __demultiplex_linearly_chunk(list chunk):
    """
    Demultiplexes a list of reads in a single-threaded approach
    """
    cdef object q = manager.Queue()

    cdef object rec = None
    # this loop will put messages in the queue containing
    # records to write
    for rec in chunk:
        demulti.demultiplex_non_perfect_record_wrapper(q, rec)
    # write matches
    __write_matches(q)


def __demultiplex_mp_chunk(list chunk):
    """
    Demultiplexes a list of reads in a multi-threaded approach
    """

    # Must use Manager queue here, or will not work
    cdef object q = manager.Queue()
    cdef object pool = mp.Pool(mp.cpu_count() - 1)

    # Fire off workers
    cdef list jobs = []
    cdef object job = None
    cdef list recs = list()
    cdef object rec = None
    cdef str read_barcode = None
    cdef object mtch = None
    cdef int i = 0
    append = jobs.append
    for rec in chunk:
        recs.append(rec)
        i += 1
        if i % mp_chunk_size == 0:
            job = pool.apply_async(demulti.demultiplex_record_wrapper, (q, recs,))
            append(job)
            recs = list()
    job = pool.apply_async(demulti.demultiplex_record_wrapper, (q, recs,))
    append(job)

    # Collect results from the workers through the pool result queue.
    for job in jobs:
        job.get()

    pool.close()
    pool.join()

    # write matches
    __write_matches(q)


def __write_matches(object q):
    """
    Processes matches in queue to write results.
    """

    cdef list mtchs = None
    cdef object mtch = None
    cdef object rec = None
    cdef str bcseq = None
    cdef object bc = None
    cdef list props = None
    cdef list tags = None
    cdef int i = 0

    #Are we really sure there are no race conditions
    #inside this loop?
    while not q.empty():

        # Extract records
        mtchs = q.get()

        for mtch in mtchs:

            # Write to info file.
            if not only_output_matched:
                f_res.write(str(mtch) + "\n")

            # No match.
            if mtch.match_type == match_type.UNMATCHED:
                if not only_output_matched:
                    re_wr.write_record(f_unmatch, mtch.record)
                    stats_total_reads_wr.increment()
                stats_unmatched.increment()
                continue

            # Append record with properties. B0:Z:Barcode, B1:Z:Prop1, B2:Z:prop3 ...
            bc = true_barcodes[mtch.barcode]
            tags = list()
            tags.append(("B0:Z", mtch.barcode))
            for i in xrange(len(bc.attributes)):
                tags.append(("B" + str(i+1) + ":Z", bc.attributes[i]))
            mtch.record.add_tags(tags)

            # Write to file.
            if mtch.match_type == match_type.MATCHED_PERFECTLY:
                re_wr.write_record(f_match, mtch.record)
                stats_perfect_matches.increment()
                stats_edit_distance_counts[0].increment()
                stats_total_reads_wr.increment()
            elif mtch.match_type == match_type.MATCHED_UNAMBIGUOUSLY:
                re_wr.write_record(f_match, mtch.record)
                stats_imperfect_unambiguous_matches.increment()
                stats_edit_distance_counts[mtch.edit_distance].increment()
                stats_total_reads_wr.increment()
            elif mtch.match_type == match_type.MATCHED_AMBIGUOUSLY:
                if not only_output_matched:
                    re_wr.write_record(f_ambig, mtch.record)
                stats_imperfect_ambiguous_matches.increment()
                stats_total_reads_wr.increment()

def print_pre_stats():
    """
    Prints pre stats
    """
    print "# Absolute output prefix path: " + outfile_prefix
    print "# Only writing matched reads: " + str(only_output_matched)
    print "# Reads format: " + re_wr.get_format()


def print_post_stats():
    """
    Prints post stats
    """
    print "# Total reads processed from infile: " \
        + str(stats_total_reads.value())
    print "# Total reads written (including multiple ambiguities): " \
        + str(stats_total_reads_wr.value())
    cdef int matched_unam = stats_perfect_matches.value() + stats_imperfect_unambiguous_matches.value()
    cdef float tot = float(stats_total_reads.value())
    print "# Reads matched unambiguously: " \
        + str(matched_unam) + " [" + str(matched_unam / tot * 100) + "%]"
    print "#    - Reads matched perfectly: " \
        + str(stats_perfect_matches.value()) + " [" + str(stats_perfect_matches.value() / tot * 100) + "%]"
    print "#    - Reads matched imperfectly: " \
        + str(stats_imperfect_unambiguous_matches.value()) \
        + " [" + str(stats_imperfect_unambiguous_matches.value() / tot * 100) + "%]"
    cdef int uniq_amb = stats_total_reads.value() - stats_perfect_matches.value() - \
                        stats_imperfect_unambiguous_matches.value() - stats_unmatched.value()
    print "# Reads matched ambiguously: " + str(uniq_amb) + " unique (making " \
        + str(stats_imperfect_ambiguous_matches.value()) + " overall) " \
        + " [" + str(uniq_amb / tot * 100) + "%]"
    print "# Reads unmatched: " \
        + str(stats_unmatched.value()) + " [" + str(stats_unmatched.value() / tot * 100) + "%]"
    cdef list distr = []
    cdef int i = 0
    for i in xrange(max_edit_distance + 1):
        distr.append(str(stats_edit_distance_counts[i].value()))
    print("# Edit distance counts for [0,...," + str(max_edit_distance) + "]: [" + ", ".join(distr) + "]")
    cdef int wall_time = time.time() - start_time
    cdef int days = wall_time / (24*60*60)
    cdef int hours = (wall_time - (24*60*60*days)) / (60*60)
    cdef int mins = (wall_time - (24*60*60*days) - (60*60*hours)) / 60
    cdef int secs = (wall_time - (24*60*60*days) - (60*60*hours) - (60*mins))
    print("# Wall time: %d:%d:%d:%d" % (days, hours, mins, secs))




"""
This package contains utilities for the Phred quality measurement.
By definition, Phred quality score Q is related to base-calling error probability P
as follows:  Q = -10 log10 P, or alternatively P = 10^{-Q/10}.
"""

import math

cpdef double probability_to_phred(double p):
    """
    Returns the Phred score of a probability.
    """
    if p < 0 or p > 1:
        raise ValueError("Probability must be in range [0,1]")
    return (-10 * math.log10(p))

cpdef double phred_to_probability(double q):
    """
    Returns the probability of a Phred score.
    """
    if q < 0:
        raise ValueError("Phred score must be in [0,inf).")
    return (math.pow(10, (-q/10)))


cpdef str sanger_phred_to_ascii(list phred):
    """
    Converts a list of sanger phred sequences to an ascii string
    :param phred: the list.
    :return: the string.
    """
    cdef list s = []
    cdef int p
    cdef int q
    for p in phred:
        q = p + 33
        if q < 33 or q > 126:
            raise ValueError("Invalid Sanger ASCII value: " + q + ". Should be in range [33,126].")
        s.append(chr(q))
    return ''.join(s)


cpdef list sanger_ascii_to_phred(str ascii_seq):
    """
    Converts an ASCII string from a SAM file or FASTQ file with Sanger Phred format
    to a list of phred scores
    """
    # Ascii range [33,126], phred range [0,93].
    cdef list phred = []
    cdef char c
    cdef int v
    for c in ascii_seq:
        v = ord(c)
        if v < 33 or v > 126:
            raise ValueError("Invalid Sanger ASCII value: " + v + ". Should be in range [33,126].")
        phred.append(v - 33)
    return phred

cpdef list sanger_ascii_to_probability(str ascii_seq):
    """
    Converts an ASCII string from a SAM file or FASTQ file with Sanger Phred format
    to a list of probabilities
    """
    cdef list probs = []
    cdef int phred
    for phred in sanger_ascii_to_phred(ascii_seq):
        probs.append(phred)
    return probs

cpdef list illumina_1_3_ascii_to_phred(str ascii_seq):
    """
    Converts an ASCII string from a Illumina 1.3-1.7 Phred format
    to a list of phred scores.
    """
    # Ascii range [64,126], phred range [0,62].
    cdef list phred = []
    cdef char c
    cdef int v
    for c in ascii_seq:
        v = ord(c)
        if v < 64 or v > 126:
            raise ValueError("Invalid Illumina 1.3 ASCII value: " + v + ". Should be in range [64,126].")
        phred.append(v - 64)
    return phred

cpdef list illumina_1_3_ascii_to_probability(str ascii_seq):
    """
    Converts an ASCII string from a Illumina 1.3-1.7 Phred format
    to a list of probabilities
    """
    cdef list probs = []
    cdef int phred
    for phred in illumina_1_3_ascii_to_phred(ascii_seq):
        probs.append(phred)
    return probs

cpdef list illumina_1_8_casava_ascii_to_phred(str ascii_seq):
    """
    Converts an ASCII string from a Illumina 1.8-x.x Phred format
    to a list of phred scores.
    """
    return sanger_ascii_to_phred(ascii_seq)

cpdef list illumina_1_8_casava_ascii_to_probability(str ascii_seq):
    """
    Converts an ASCII string from a Illumina 1.8-x.x Phred format
    to a list of probabilities
    """
    return sanger_ascii_to_probability(ascii_seq)

cpdef list solid_ascii_to_phred(str ascii_seq):
    """
    Converts an ASCII string from a SOLiD Phred format
    to a list of phred scores.
    """
    return sanger_ascii_to_phred(ascii_seq)

cpdef list solid_ascii_to_probability(str ascii_seq):
    """
    Converts an ASCII string from a SOLiD Phred format
    to a list of probabilities
    """
    return sanger_ascii_to_probability(ascii_seq)





"""
Defines different kinds of matches.
"""
cdef int KILL = -1
cdef int UNMATCHED = 0
cdef int MATCHED_PERFECTLY = 1
cdef int MATCHED_UNAMBIGUOUSLY = 2
cdef int MATCHED_AMBIGUOUSLY = 3

cdef str match_type_to_str(int match_type):
    """
    Simple converter from match type to string.
    """
    if match_type == 0:
        return "UNMATCHED"
    elif match_type == 1:
        return "MATCHED_PERFECTLY"
    elif match_type == 2:
        return "MATCHED_UNAMBIGUOUSLY"
    elif match_type == 3:
        return "MATCHED_AMBIGUOUSLY"
    return -1


from setuptools import find_packages
from glob import glob
import numpy
from distutils.core import setup
from distutils.extension import Extension
try:
    from Cython.Distutils import build_ext
except ImportError:
    use_cython = False
else:
    use_cython = True
#from Cython.Distutils.extension import Extension
#from Cython.Distutils import build_ext
#from Cython.Build import cythonize


cmdclass = { }
ext_modules = [ ]


if use_cython:
    ext_modules += [
    Extension("taggd.core.demultiplex_core_functions",   ["taggd/core/demultiplex_core_functions.pyx"]),
    Extension("taggd.core.demultiplex_record_functions", ["taggd/core/demultiplex_record_functions.pyx"]),
    Extension("taggd.core.demultiplex_search_functions", ["taggd/core/demultiplex_search_functions.pyx"]),
    Extension("taggd.core.match",                        ["taggd/core/match.pyx"]),
    Extension("taggd.core.match_type",                   ["taggd/core/match_type.pyx"]),
    Extension("taggd.misc.distance_metrics",             ["taggd/misc/distance_metrics.pyx"]),
    Extension("taggd.misc.kmer_utils",                   ["taggd/misc/kmer_utils.pyx"]),
	Extension("taggd.misc.counter",                      ["taggd/misc/counter.pyx"]),
	Extension("taggd.io.phred_utils",                    ["taggd/io/phred_utils.pyx"]),
    Extension("taggd.io.fastq_utils",                    ["taggd/io/fastq_utils.pyx"]),
    Extension("taggd.io.barcode_utils",                  ["taggd/io/barcode_utils.pyx"]),
    Extension("taggd.io.record",                         ["taggd/io/record.pyx"]),
    Extension("taggd.io.sam_record",                     ["taggd/io/sam_record.pyx"]),
    Extension("taggd.io.fasta_record",                   ["taggd/io/fasta_record.pyx"]),
    Extension("taggd.io.fastq_record",                   ["taggd/io/fastq_record.pyx"]),
    Extension("taggd.io.reads_reader_writer",            ["taggd/io/reads_reader_writer.pyx"])
    ]
    cmdclass.update({ 'build_ext': build_ext })
else:
    ext_modules += [
    Extension("taggd.core.demultiplex_core_functions",   ["taggd/core/demultiplex_core_functions.c"]),
    Extension("taggd.core.demultiplex_record_functions", ["taggd/core/demultiplex_record_functions.c"]),
    Extension("taggd.core.demultiplex_search_functions", ["taggd/core/demultiplex_search_functions.c"]),
    Extension("taggd.core.match",                        ["taggd/core/match.c"]),
    Extension("taggd.core.match_type",                   ["taggd/core/match_type.c"]),
    Extension("taggd.misc.distance_metrics",             ["taggd/misc/distance_metrics.c"]),
    Extension("taggd.misc.kmer_utils",                   ["taggd/misc/kmer_utils.c"]),
	Extension("taggd.misc.counter",                      ["taggd/misc/counter.c"]),
	Extension("taggd.io.phred_utils",                    ["taggd/io/phred_utils.c"]),
    Extension("taggd.io.fastq_utils",                    ["taggd/io/fastq_utils.c"]),
    Extension("taggd.io.barcode_utils",                  ["taggd/io/barcode_utils.c"]),
    Extension("taggd.io.record",                         ["taggd/io/record.c"]),
    Extension("taggd.io.sam_record",                     ["taggd/io/sam_record.c"]),
    Extension("taggd.io.fasta_record",                   ["taggd/io/fasta_record.c"]),
    Extension("taggd.io.fastq_record",                   ["taggd/io/fastq_record.c"]),
    Extension("taggd.io.reads_reader_writer",            ["taggd/io/reads_reader_writer.c"])
    ]


setup(
	name="taggd",
	version = '0.1.10',
	author = 'Joel Sjostrand',
	author_email = 'joel.sjostrand@scilifelab.se',
	license = 'Open BSD',
    description = 'Bioinformatics genetic barcode demultiplexing',
    url = 'https://github.com/JoelSjostrand/taggd',
    download_url = 'https://github.com/JoelSjostrand/taggd/0.1.10',
	scripts = glob("scripts/*.py"),
    packages = ['taggd', 'taggd.core', 'taggd.io', 'taggd.misc'],
    package_data = {'': ['*.pyx', '*.pxd', '*.h', '*.c'], },
	install_requires = [
	    'setuptools',
	    'pysam',
	   	'numpy',
	],
	test_suite = 'tests',
	cmdclass = cmdclass,
    ext_modules=ext_modules,
	include_dirs = [numpy.get_include(), '.'],
    keywords = ['bioinformatics', 'demultiplexing']
    )



"""
This represent a barcode match with all
the necessary information that can be used to
write to a file later
"""
cimport taggd.core.match_type as match_type

cdef str get_match_header():
    """
    Returns a header for the match-to-string conversion.
    """
    return "#Annotation\tMatch_result\tBarcode\tEdit_distance\tAmbiguous_top_hits\t" \
        "Qualified_candidates\tRaw_candidates\tLast_position\tApprox_insertions\tApprox_deletions"

class Match(object):
    """
    Shorthand for a match.
    """
    def __init__(self,
                 object record_,
                 int match_type_,
                 str barcode_,
                 int edit_distance_,
                 int ambiguous_top_hits_,
                 int qualified_candidates_,
                 int raw_candidates_,
                 int last_position_,
                 int insertions_read_,
                 int insertions_barcode_):
        """
        Constructor.
        """
        self.record = record_
        self.match_type = match_type_
        self.barcode = barcode_
        self.edit_distance = edit_distance_
        self.ambiguous_top_hits = ambiguous_top_hits_
        self.qualified_candidates = qualified_candidates_
        self.raw_candidates = raw_candidates_
        self.last_position = last_position_
        self.insertions_read = insertions_read_
        self.insertions_barcode = insertions_barcode_


    def __str__(self):
        """
        String converter.
        TODO this is slow, python recommends to build strings
        in the form ''.{} ....
        """
        return (self.record.annotation + '\t' +
                match_type.match_type_to_str(self.match_type) + '\t' +
                self.barcode + '\t' +
                str(self.edit_distance) + '\t' +
                str(self.ambiguous_top_hits) + '\t' +
                str(self.qualified_candidates) + '\t' +
                str(self.raw_candidates) + '\t' +
                str(self.last_position) + '\t' +
                str(self.insertions_read) + '\t' +
                str(self.insertions_barcode))


""" 
This class contains the main interface of taggd
Parameters will be captured, parsed and validated. 
The main workflow will be executed and output files
will be generated
"""
import os
import argparse
import taggd.io.barcode_utils as bu
import taggd.core.demultiplex_core_functions as core
import taggd.core.demultiplex_record_functions as rec
import taggd.core.demultiplex_search_functions as srch

def main(argv=None):
    """Main application."""

    # Create a parser
    parser = argparse.ArgumentParser(description=__doc__)

    # Needed parameters
    parser.add_argument('barcodes_infile', 
                        metavar='barcodes-infile', 
                        help="The file with true barcode IDs and other properties.")
    parser.add_argument('reads_infile', 
                        metavar='reads-infile', 
                        help="The FASTQ, FASTA, SAM or BAM file with reads.")
    parser.add_argument('outfile_prefix', 
                        metavar='outfile-prefix', help="The output files prefix.")

    # Optional arguments.
    parser.add_argument('--start-position', 
                        type=int,
                         help='The start position for barcodes in reads (default: %(default)d)', 
                         default=0, metavar="[int]")
    parser.add_argument('--k', 
                        type=int, 
                        help='The kmer length (default: %(default)d)', 
                        default=7, metavar="[int]")
    parser.add_argument('--max-edit-distance', 
                        type=int, 
                        help='The max edit distance for allowing hits (default: %(default)d)', 
                        default=5, metavar="[int]")
    parser.add_argument('--metric', 
                        help= "Distance metric: Subglobal, Levenshtein or Hamming (default: %(default)s)", 
                        default="Subglobal", metavar="[string]")
    parser.add_argument('--slider-increment', 
                        type=int, help="Space between kmer searches, " \
                        "0 yields kmer length (default: %(default)d)", 
                        default=0, metavar="[int]")
    parser.add_argument('--overhang', 
                        type=int, 
                        help="Additional flanking bases around read barcode " \
                        "to allow for insertions (default: %(default)d)", 
                        default=2, metavar="[int]")
    parser.add_argument('--only-output-matched', 
                        help="Suppresses writing of output to only file with matched reads.", 
                        default=False, action='store_true')
    parser.add_argument('--seed', 
                        help="Random number generator seed for shuffling ambiguous hits (default: %(default)s)", 
                        default=None, metavar="[string]")
    parser.add_argument('--no-multiprocessing',
                         help="If set, turns off multiprocessing of reads", 
                         default=False, action='store_true')
    parser.add_argument('--estimate-min-edit-distance', 
                        type=int, 
                        help="If set, estimates the min edit distance among true " \
                        "barcodes by comparing the specified number of pairs, " \
                        "0 means no estimation (default: %(default)d)", 
                        default=0, metavar="[int]")
    parser.add_argument('--chunk-size',
                        type=int, 
                        help="Chunk of maximum number of simultaneously " \
                        "processed reads (default: %(default)d)", 
                        default=50000, metavar="[int]")
    parser.add_argument('--mp-chunk-size',
                        type=int,
                        help="Chunk of maximum number of " \
                        "processed reads sent to each thread (default: %(default)d)",
                        default=500, metavar="[int]")
    parser.add_argument('--no-offset-speedup', 
                        help="Turns off an offset speedup routine, " \
                        "increasing time but may yield more hits.", 
                        default=False, action='store_true')

    # Parse
    if argv == None:
        options = parser.parse_args()
    else:
        options = parser.parse_args(argv)

    # Validate all options.
    if not os.path.isfile(options.barcodes_infile) :
        raise ValueError("Invalid true barcodes input file path.")
    if not os.path.isfile(options.reads_infile) :
        raise ValueError("Invalid reads input file path.")
    if not (options.reads_infile.upper().endswith(".FASTQ") or \
            options.reads_infile.upper().endswith(".FQ") or \
            options.reads_infile.upper().endswith(".SAM") or \
            options.reads_infile.upper().endswith(".FASTA") or \
            options.reads_infile.upper().endswith(".FA") or \
            options.reads_infile.upper().endswith(".BAM")):
        raise ValueError("Invalid reads input file format: must be FASTQ, " \
                         "FASTA, SAM or BAM format and file end with .fq, fastq, .fa, .fasta, .sam or .bam")
    if options.outfile_prefix is None or options.outfile_prefix == "":
        raise ValueError("Invalid output file prefix.")
    if options.k <= 0:
        raise ValueError("Invalid kmer length. Must be > 0.")
    if options.max_edit_distance < 0:
        raise ValueError("Invalid max edit distance. Must be >= 0.")
    if options.metric not in ("Subglobal", "Levenshtein", "Hamming"):
        raise ValueError("Invalid metric. Must be Subglobal, Levenshtein or Hamming.")
    if options.slider_increment < 0:
        raise ValueError("Invalid slider increment. Must be >= 0.")
    if options.slider_increment == 0:
        options.slider_increment = int(options.k)
    if options.start_position < 0:
        raise ValueError("Invalid start position. Must be >= 0.")
    if options.overhang < 0:
        raise ValueError("Invalid overhang. Must be >= 0.")
    if options.metric == "Hamming" and options.overhang > 0:
        raise ValueError("Invalid overhang. Must be 0 for Hamming metric.")
    if options.chunk_size <= 0:
        raise ValueError("Invalid max chunk size. Must be > 0.")
    if options.mp_chunk_size <= 0:
        raise ValueError("Invalid multiprocessing max chunk size. Must be > 0.")

    # Read barcodes file
    true_barcodes = bu.read_barcode_file(options.barcodes_infile)

    # Check barcodes file.
    if options.estimate_min_edit_distance > 0:
        min_dist = bu.estimate_min_edit_distance(true_barcodes, options.estimate_min_edit_distance)
        if min_dist <= options.max_edit_distance:
            raise ValueError("Invalid max edit distance: exceeds or equal " \
                             "to estimated minimum edit distance among true barcodes.")
        print "# Minimum edit distance between true barcodes was estimated to (may be less) " + str(min_dist)

    # Initialize main components
    core.init(true_barcodes,
              options.reads_infile,
              os.path.abspath(options.outfile_prefix),
              options.start_position,
              options.max_edit_distance,
              options.no_multiprocessing,
              options.only_output_matched,
              options.chunk_size,
              options.mp_chunk_size)

    rec.init(true_barcodes,
             options.start_position,
             min(options.start_position, options.overhang),
             options.overhang,
             options.max_edit_distance,
             options.seed)

    srch.init(true_barcodes,
              options.k,
              options.max_edit_distance,
              options.metric,
              options.slider_increment, 
              min(options.start_position, options.overhang), 
              options.overhang,
              options.no_offset_speedup)

    # Demultiplex
    core.print_pre_stats()
    rec.print_pre_stats()
    print "# Starting demultiplexing with the following options..." + str(options)
    core.demultiplex()
    print "# ...finished demultiplexing"
    core.print_post_stats()


"""
Contains functions for demultiplexing a record.
"""

import taggd.core.match as match
cimport taggd.core.match as match
import taggd.core.match_type as match_type
cimport taggd.core.match_type as match_type
import taggd.core.demultiplex_search_functions as srch
cimport taggd.core.demultiplex_search_functions as srch
import random
from cpython cimport bool

# User settings.
cdef dict true_barcodes
cdef int start_position
cdef int barcode_length
cdef int pre_overhang
cdef int post_overhang
cdef int max_edit_distance
cdef str seed


def init(dict true_barcodes_,
         int start_position_,
         int pre_overhang_,
         int post_overhang_,
         int max_edit_distance_,
         str seed_):
    """
    Initializes variables for matching.
    :param true_barcodes_: barcodes dict to attributes.
    :param start_position_: the start position.
    :param pre_overhang_: how many flanking bases pre for indels. Not applicable for Hamming distance.
    :param post_overhang_: how many flanking bases post for indels. Not applicable for Hamming distance.
    :param max_edit_distance_: max allowed distance.
    :param seed_: PRNG seed.
    """

    global true_barcodes
    true_barcodes = true_barcodes_

    # User settings.
    global start_position
    start_position = start_position_
    global pre_overhang
    pre_overhang = pre_overhang_
    global post_overhang
    post_overhang = post_overhang_
    global barcode_length
    barcode_length = len(true_barcodes.keys()[0])
    global max_edit_distance
    max_edit_distance = max_edit_distance_
    global seed
    seed = seed_
    random.seed(seed)


def demultiplex_record_wrapper(object q, object rec):
    """
    Non cdef wrapper for cdef:ed multithreading function.
    """
    return demultiplex_record(q, rec)



cdef bool demultiplex_record(object q, object recs):
    """
    Demultiplexes a record. Init must have been called before.
    :param q: the queue where results are stored.
    :param recs: the records.
    :return: true if matched.
    """

    cdef str read_barcode = None
    cdef dict candidates = None
    cdef list qual_hits = None
    cdef list top_hits = None
    cdef str bcseq = None
    cdef int dist = 0
    cdef int last_pos = -1
    cdef int a = -1
    cdef int b = -1
    cdef object mtch = None
    cdef list mtchs = list()

    # Iterate over records.
    for rec in recs:

        # Try perfect hit first.
        read_barcode = rec.sequence[start_position:(start_position+barcode_length)]
        if read_barcode in true_barcodes:
            mtch = match.Match(rec, match_type.MATCHED_PERFECTLY, read_barcode, 0, 1, 1, -1, barcode_length-1, 0, 0)
            mtchs.append(mtch)
            continue

        # Include overhang.
        read_barcode = rec.sequence[(start_position - pre_overhang):min(len(rec.sequence), \
                            (start_position + barcode_length + post_overhang))]

        # Narrow down hits.
        candidates = srch.get_candidates(read_barcode)
        qual_hits = srch.get_distances(read_barcode, candidates)
        top_hits = srch.get_top_hits(qual_hits)

        if not top_hits:
            # NO MATCH.
            # Record   Match type   Barcode
            # [Edit distance, Ambiguous top hits, Qualified candidates,
            # Raw candidates, Last_pos, Insertions read, Insertions true]
            mtch = match.Match(rec, match_type.UNMATCHED, "-", -1, 0, len(qual_hits), len(candidates), -1, -1, -1)
            mtchs.append(mtch)
        else:
            if len(top_hits) == 1:
                # SINGLE MATCH.
                bcseq, dist, last_pos, a, b = top_hits[0]
                # Record   Match type   Barcode
                # [Edit distance, Ambiguous top hits, Qualified candidates,
                # Raw candidates, Last_pos, Insertions read, Insertions true]
                mtch = match.Match(rec, match_type.MATCHED_UNAMBIGUOUSLY, bcseq, dist, len(top_hits), len(qual_hits),
                                   len(candidates), last_pos, a, b)
                mtchs.append(mtch)
            else:
                # AMBIGUOUS MATCHES.
                # Multiple best hits. Shuffle results to avoid biases.
                random.shuffle(top_hits)
                for (bcseq, dist, last_pos, a, b) in top_hits:
                    # Record   Match type   Barcode
                    # [Edit distance, Ambiguous top hits, Qualified candidates,
                    # Raw candidates, Last_pos, Insertions read, Insertions true]
                    mtch = match.Match(rec, match_type.MATCHED_AMBIGUOUSLY, bcseq, dist, len(top_hits),
                            len(qual_hits), len(candidates), last_pos, a, b)
                    mtchs.append(mtch)
    q.put(mtchs)
    return True



def print_pre_stats():
    """
    Prints pre stats
    """
    print "# True barcodes length: " + str(barcode_length)
    print "# Read barcodes length when overhang added: " + str(barcode_length + pre_overhang + post_overhang)